/* --------------------------------------------------------------------------
 *
 * License
 *
 * The contents of this file are subject to the Jabber Open Source License
 * Version 1.0 (the "License").  You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License.  You
 * may obtain a copy of the License at http://www.jabber.com/license/ or at
 * http://www.opensource.org/.  
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyrights
 * 
 * Portions created by or assigned to Jabber.com, Inc. are 
 * Copyright (c) 1999-2000 Jabber.com, Inc.  All Rights Reserved.  Contact
 * information for Jabber.com, Inc. is available at http://www.jabber.com/.
 *
 * Portions Copyright (c) 1998-1999 Jeremie Miller.
 * 
 * Acknowledgements
 * 
 * Special thanks to the Jabber Open Source Contributors for their
 * suggestions and support of Jabber.
 * 
 * --------------------------------------------------------------------------*/
 
#include <jabberd.h>


/* 
gcc -fPIC -shared -o xdb_file.so xdb_file.c -I../src

needs to have a spool section in the config:

<load><dispatch>../load/dispatch.so</dispatch></load>
<dispatch xmlns="jabber:config:dispatch">
  <domain>icq</domain>
  <random/>
  <record>machine1</record>
  <record>machine2</record>
  <record>machine3</record>
  <record>machine4</record>
</dispatch>

within the spool, xdb_file will make folders for hostnames it has to save data for, and within those save username.xml files containing a user's namespaces

*/

#define RR      "roundrobin"
#define GREEDY  "greedy"
#define RANDOM  "random"

#define HASHBUF_LEN 41
#define MAX_SESSIONS 2003
typedef enum { t_ROUNDROBIN, t_GREEDY, t_RANDOM } typedispatch;


typedef struct __user_session{
    pool p;
    int numserver;
} *user_session, _user_session;


typedef struct subserver_struct
{
    char * name;
    int nconnexion;
    int load;
} *subserver, _subserver;


typedef struct dispatch_struct
{
    char * name;
    typedispatch type;
    int nsubserver;
    subserver subservers;
    HASHTABLE sessions;
    int nrr;
} *dispatcher, _dispatcher;


result dispatch_phandler(instance i, dpacket p, void *arg){
    char * to = NULL;
    char * from = NULL;
    char hashbuf[HASHBUF_LEN];
    int j, accu;
    dispatcher dsp = (dispatcher) arg;
    jid jid_from;
    


    if ((to = xmlnode_get_attrib(p->x, "to")) == NULL ||
	(from = xmlnode_get_attrib(p->x, "from")) == NULL ||
	dsp == NULL)
	return r_ERR;


    fprintf(stderr, "++++> jid_full : %s\n", from);    

    switch(dsp->type){


    case t_RANDOM:
	memset(hashbuf, 0, HASHBUF_LEN);
	shahash_r(from, hashbuf);
	
	for (accu = 0, j = 0 ; j < HASHBUF_LEN ; j++){
	    accu += hashbuf[j];
	}

	accu = accu % dsp->nsubserver;
	fprintf(stderr, "++> server : %s\n", dsp->subservers[accu].name);
	
	jid_from = jid_new(i->p, to);
	fprintf(stderr, "++> server : %s\n", dsp->subservers[accu].name);
	fprintf(stderr, "++> server : %s\n", dsp->subservers[accu].name);
	fprintf(stderr, "++> server : %s\n", dsp->subservers[accu].name);
	jid_set(jid_from, pstrdup(xmlnode_pool(p->x), dsp->subservers[accu].name), JID_SERVER);
	dsp->subservers[accu].nconnexion++;
	xmlnode_put_attrib(p->x, "to", jid_full(jid_from));

	fprintf(stderr, "++> Sending: %s\n", xmlnode2str(p->x));

	deliver(dpacket_new(p->x), NULL);
	return r_DONE;

	break;

    case t_GREEDY:
    {
	user_session session = NULL;

	if ((session = (user_session) ghash_get(dsp->sessions, from)) == NULL){
	    pool po;
	    int min;
	    int k;
	    int choosen;
	    
	    po = pool_new();
	    session = (user_session) pmalloco(po, sizeof(user_session));
	    session->p = po;

	    for (choosen = 0, min = dsp->subservers[0].nconnexion, k = 0 ; k < dsp->nsubserver ; k++){
		if (dsp->subservers[k].nconnexion < min){
		    choosen = k;
		    min = dsp->subservers[k].nconnexion;
		}
	    }

	    session->numserver = k;
	    ghash_put(dsp->sessions, pstrdup(po, from), session);
	    dsp->subservers[k].nconnexion++;
	}

	jid_from = jid_new(i->p, to);
	jid_set(jid_from, pstrdup(xmlnode_pool(p->x), dsp->subservers[session->numserver].name), JID_SERVER);
	xmlnode_put_attrib(p->x, "to", jid_full(jid_from));
	fprintf(stderr, "++> Sending: %s\n", xmlnode2str(p->x));


	if (xmlnode_get_tag(p->x, "presence") &&
	    !j_strcmp(xmlnode_get_attrib(p->x, "type"), "unavailable")){
	    
	    if ((session = (user_session) ghash_get(dsp->sessions, from)) != NULL){
		ghash_remove(dsp->sessions, session);
		pool_free(session->p);
	    }
	}

	deliver(dpacket_new(p->x), NULL);

	return r_DONE;
    }
	
    case t_ROUNDROBIN:
    {
	user_session session = NULL;

	fprintf(stderr, "ROUDROBIN\n");


	if ((session = (user_session) ghash_get(dsp->sessions, from)) == NULL){
	    pool po;
	    
	    po = pool_new();
	    session = (user_session) pmalloco(po, sizeof(user_session));
	    session->p = po;

	    session->numserver = dsp->nrr;
	    dsp->subservers[dsp->nrr].nconnexion++;
	    if (dsp->nrr < dsp->subservers)
		dsp->nrr++;
	    else
		dsp = 0;

	    ghash_put(dsp->sessions, pstrdup(po, from), session);
	}

	jid_from = jid_new(i->p, to);
	jid_set(jid_from, pstrdup(xmlnode_pool(p->x), dsp->subservers[session->numserver].name), JID_SERVER);

	xmlnode_put_attrib(p->x, "to", jid_full(jid_from));
	fprintf(stderr, "++> Sending: %s\n", xmlnode2str(p->x));


	if (xmlnode_get_tag(p->x, "presence") &&
	    !j_strcmp(xmlnode_get_attrib(p->x, "type"), "unavailable")){
	    
	    if ((session = (user_session) ghash_get(dsp->sessions, from)) != NULL){
		ghash_remove(dsp->sessions, session);
		pool_free(session->p);
	    }
	}

	deliver(dpacket_new(p->x), NULL);

	return r_DONE;
	break;
    }


    default:
	break;

    }		    

    return r_ERR;
}


void dispatch(instance i, xmlnode x)
{
    xmlnode config;
    xdbcache xc;
    dispatcher dsp;
    xmlnode ch;
    int nrecord, j;
    
    log_debug(ZONE,"dispatch loading");

    xc = xdb_cache(i);
    config = xdb_get(xc, jid_new(xmlnode_pool(x),"config@-internal"),"jabber:config:dispatch");
    
    dsp = pmalloco(i->p, sizeof(_dispatcher));

    dsp->name = xmlnode_get_data(xmlnode_get_tag(config, "domain"));
    if (xmlnode_get_tag(config, RR))
	dsp->type = t_ROUNDROBIN;
    else if (xmlnode_get_tag(config, GREEDY))
	dsp->type = t_GREEDY;
    else if (xmlnode_get_tag(config, RANDOM))
	dsp->type = t_RANDOM;

    for(nrecord = 0, ch = xmlnode_get_firstchild(config); ch != NULL; ch = xmlnode_get_nextsibling(ch))
    {
        if (xmlnode_get_type(ch) != NTYPE_TAG)
            continue;

        if (j_strcmp(xmlnode_get_name(ch),"record") == 0)
	    nrecord++;
    }

    dsp->nsubserver = nrecord;
    dsp->subservers = pmalloco(i->p, nrecord*sizeof(_subserver));
    dsp->nrr = 0;
    
    for(j = 0, ch = xmlnode_get_firstchild(config); ch != NULL; ch = xmlnode_get_nextsibling(ch))
    {
        if (xmlnode_get_type(ch) != NTYPE_TAG)
            continue;

        if (j_strcmp(xmlnode_get_name(ch), "record") == 0){
	    dsp->subservers[j].name = pstrdup(i->p, xmlnode_get_data(ch));
	    j++;
	}
    }


    if (dsp->type == t_ROUNDROBIN ||
	dsp->type == t_GREEDY){
	dsp->sessions = ghash_create(MAX_SESSIONS, (KEYHASHFUNC)str_hash_code, (KEYCOMPAREFUNC)j_strcmp);
    }

    register_phandler(i, o_DELIVER, dispatch_phandler, (void *)dsp);

    xmlnode_free(config);
//    pool_cleanup(i->p, xdb_file_cleanup, (void*)xf);
}




